# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import DataProcessInfoClass
from .....schema_classes import DataProcessInstanceInputClass
from .....schema_classes import DataProcessInstanceOutputClass
from .....schema_classes import DataProcessInstancePropertiesClass
from .....schema_classes import DataProcessInstanceRelationshipsClass
from .....schema_classes import DataProcessInstanceRunEventClass
from .....schema_classes import DataProcessInstanceRunResultClass
from .....schema_classes import DataProcessRunStatusClass
from .....schema_classes import DataProcessTypeClass
from .....schema_classes import RunResultTypeClass


DataProcessInfo = DataProcessInfoClass
DataProcessInstanceInput = DataProcessInstanceInputClass
DataProcessInstanceOutput = DataProcessInstanceOutputClass
DataProcessInstanceProperties = DataProcessInstancePropertiesClass
DataProcessInstanceRelationships = DataProcessInstanceRelationshipsClass
DataProcessInstanceRunEvent = DataProcessInstanceRunEventClass
DataProcessInstanceRunResult = DataProcessInstanceRunResultClass
DataProcessRunStatus = DataProcessRunStatusClass
DataProcessType = DataProcessTypeClass
RunResultType = RunResultTypeClass

# fmt: on
