{
  "type": "record",
  "name": "DataProcessInstanceProperties",
  "namespace": "com.linkedin.pegasus2avro.dataprocess",
  "doc": "The inputs and outputs of this data process",
  "fields": [
    {
      "name": "customProperties",
      "type": {
        "type": "map",
        "values": "string"
      },
      "doc": "Custom property bag.",
      "default": {},
      "Searchable": {
        "/*": {
          "queryByDefault": true
        }
      }
    },
    {
      "name": "externalUrl",
      "type": [
        "null",
        "string"
      ],
      "doc": "URL where the reference exist",
      "default": null,
      "java": {
        "class": "com.linkedin.pegasus2avro.common.url.Url",
        "coercerClass": "com.linkedin.pegasus2avro.common.url.UrlCoercer"
      }
    },
    {
      "name": "name",
      "type": "string",
      "doc": "Process name",
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "type",
      "type": [
        "null",
        {
          "type": "enum",
          "name": "DataProcessType",
          "symbols": [
            "BATCH_SCHEDULED",
            "BATCH_AD_HOC",
            "STREAMING"
          ]
        }
      ],
      "doc": "Process type",
      "default": null,
      "Searchable": {
        "addToFilters": true,
        "fieldType": "KEYWORD",
        "filterNameOverride": "Process Type"
      }
    },
    {
      "name": "created",
      "type": {
        "type": "record",
        "name": "AuditStamp",
        "namespace": "com.linkedin.pegasus2avro.common",
        "doc": "Data captured on a resource/association/sub-resource level giving insight into when that resource/association/sub-resource moved into a particular lifecycle stage, and who acted to move it into that specific lifecycle stage.",
        "fields": [
          {
            "name": "time",
            "type": "long",
            "doc": "When did the resource/association/sub-resource move into the specific lifecycle stage represented by this AuditEvent."
          },
          {
            "name": "actor",
            "type": "string",
            "doc": "The entity (e.g. a member URN) which will be credited for moving the resource/association/sub-resource into the specific lifecycle stage. It is also the one used to authorize the change.",
            "java": {
              "class": "com.linkedin.pegasus2avro.common.urn.Urn"
            }
          },
          {
            "name": "impersonator",
            "type": [
              "null",
              "string"
            ],
            "doc": "The entity (e.g. a service URN) which performs the change on behalf of the Actor and must be authorized to act as the Actor.",
            "default": null,
            "java": {
              "class": "com.linkedin.pegasus2avro.common.urn.Urn"
            }
          }
        ]
      },
      "doc": "Audit stamp containing who reported the lineage and when",
      "Searchable": {
        "/time": {
          "fieldName": "created",
          "fieldType": "COUNT",
          "queryByDefault": false
        }
      }
    }
  ],
  "Aspect": {
    "name": "dataProcessInstanceProperties"
  }
}